/* replace.c - The ideal program for sed-haters (i.e. everyone).
   (C) Richard K. Lloyd 1998

   Provides a simple way of replacing strings in lines without resorting to
   the ludicrous gibberish of sed, whose syntax was obviously dreamt up by
   someone on LSD !

   Syntax:
   replace [qualifiers] <oldstr> <newstr> [-a <oldstr> <newstr> [-a...]]
           [filename...]

   What "replace" does by default:

   It performs a case-insensitive search for the "word" <oldstr> in each text
   line and replaces it with <newstr>, which is case-adjusted to match the
   case of the particular occurrence of <oldstr> in the original line. A "word"
   is defined to be a string that is preceded AND followed by a non-
   alphanumeric character (start and end of lines count as such a character)
   in the original text line. Multiple strings can be replaced in order using
   the "-a" qualifier.

   The qualifiers modify this behaviour appropriately (try "replace -?"). */
  
#include "replace.h"
#include "patchlevel.h"

static char *oldstr[MAXSTRS+1];
static char *newstr[MAXSTRS+1];
static char intitle[MAXSTRLEN];
static char progname[MAXSTRLEN];
static int oldstrlen[MAXSTRS],newstrlen[MAXSTRS];
static int numstrs=0; /* No strings yet */

/* Default settings if no qualifiers supplied */
static int sensitive=0; /* Case-insensitive */
static int word=1; /* Match words only, not sub-strings */
static int force=0; /* Don't force it - keep .cln version */
static int padit=0; /* Don't pad new string */
static int verbose=0; /* Not verbose */
static int startcol=1; /* Start replacing from column 1 */
static int startline=1; /* Start replacing from line 1 */

static int updated;
static int linenum;
static int repcount;

static void tidy_up()
{
   /* I don't actually have to have this routine since UNIX does it for me,
      but let's be careful anyway */
   int f;
   for (f=1;f<=numstrs;f++)
   {
      (void)free((MALLOCPTR)oldstr[f]);
      (void)free((MALLOCPTR)newstr[f]);
   }
}

static void leave(errmess)
char *errmess;
{
   tidy_up();
   if (errmess[0]) (void)fprintf(stderr,"%s: %s\n",progname,errmess);
   (void)exit(1);
}

char *basename(fullname)
char *fullname;
{
    char *slashptr;
    if ((slashptr=strrchr(fullname,'/'))==(char *)NULL)
       return(fullname);
    else
       return(&slashptr[1]);
}

static char *replace_line(oldline,strnum)
char *oldline;
int strnum;
{
   static char newline[MAXSTRLEN];
   int oldlen=strlen(oldline);
   size_t osl=oldstrlen[strnum],nsl=newstrlen[strnum];
   char *ostr=oldstr[strnum],*nstr=newstr[strnum];
   int oldmax=oldlen-osl;
   int a=0,b=0,c,comppos,truncated=0;
   char newchar;
   if (linenum>=startline)
   {
      while (a<startcol && a<oldlen && !truncated)
      { COPY_CHAR(oldline[a++]); }
      while (a<=oldmax && !truncated)
      {
         if (sensitive)
            c=strncmp(&oldline[a],ostr,osl);
         else
            c=strncasecmp(&oldline[a],ostr,osl);
         if (word && !c)
         {
            if (a) c=isalnum(oldline[a-1]);
            if (!c && a<oldmax) c=isalnum(oldline[a+osl]);
         }
         if (c) { COPY_CHAR(oldline[a++]); }
         else
         {
            comppos=a;
            for (c=0;(c<nsl) & !truncated;c++)
            {
               newchar=nstr[c];
               if (!sensitive)
               {
                  if (isupper(oldline[comppos++]))
                     newchar=toupper(newchar);
                  else newchar=tolower(newchar);
                  if (c>=osl-1) comppos--;
               }
               COPY_CHAR(newchar);
            }
            if (verbose)
            {
               char tmpold[MAXSTRLEN],tmpnew[MAXSTRLEN];
               (void)strncpy(tmpold,&oldline[a],osl); tmpold[osl]='\0';
               (void)strncpy(tmpnew,&newline[b-nsl],nsl); tmpnew[nsl]='\0';
               (void)fprintf(stderr,"\"%s\" -> \"%s\" (pos: %d, line: %d, file: %s)\n",
                             tmpold,tmpnew,a+1,linenum,intitle);
            }
            repcount++;
            a+=osl;
         }
      }
   }
   while (a<oldlen && !truncated) { COPY_CHAR(oldline[a++]); }
   newline[b]='\0';
   if (truncated)
      (void)fprintf(stderr,"WARNING: Line %d was truncated due to excessive length\n",linenum);
   return(newline);
}

static void plural(str,value)
char *str;
int value;
{
   switch (value)
   {
      case 0: (void)fprintf(stderr,"no %ss ",str); break;
      case 1: (void)fprintf(stderr,"one %s ",str); break;
      case 2: (void)fprintf(stderr,"two %ss ",str); break;
      case 3: (void)fprintf(stderr,"three %ss ",str); break;
      default: (void)fprintf(stderr,"%d %ss ",value,str);
   } 
}

static void usage()
/* Tell the user what to type next time... */
{
   (void)fprintf(stderr,"\nreplace V%s - The sane person's alternative to sed (C) Richard K. Lloyd 1998\n",REPLACE_VERSION);
   (void)fprintf(stderr,"A case-insensitive word searcher and substituter designed for sed-haters.\n\n");
   (void)fprintf(stderr,"Syntax: replace [-?] [-c startcol] [-e] [-l linenum] [-s] [-v]\n");
   (void)fprintf(stderr,"                old_str new_str [-a old_str new_str [-a...]]\n");
   (void)fprintf(stderr,"                [filename...]\n\n");
   (void)fprintf(stderr,"-? displays this syntax message.\n");
   (void)fprintf(stderr,"-a allows extra pairs of strings to be replaced in order.\n");
   (void)fprintf(stderr,"-c startcol starts the string replace from column 'startcol' rather than the\n");
   (void)fprintf(stderr,"   first column.\n");
   (void)fprintf(stderr,"-e makes search case-sensitive. new_str exactly replaces old_str\n");
   (void)fprintf(stderr,"   with no case-adjustment to new_str.\n");
   (void)fprintf(stderr,"-f forces the update of files without any .cln backup.\n");
   (void)fprintf(stderr,"-l linenum starts the string replacement from line 'linenum' rather than the\n");
   (void)fprintf(stderr,"   first line.\n");
   (void)fprintf(stderr,"-p pads new_str with trailing spaces if it is shorter than old_str.\n");
   (void)fprintf(stderr,"-s relaxes the condition that old_str has to match a 'word' i.e. it replaces\n");
   (void)fprintf(stderr,"   old_str even if it is a substring of a 'word'.\n");
   (void)fprintf(stderr,"-v switches on verbose mode, reporting all replacements to stderr.\n");
   leave("");
}

static void get_options(argc,argv)
int argc;
char **argv;
/* Parse the standard qualifiers */
{
   int param,alloclen,newtmp,oldtmp,looping=1;

   (void)strcpy(progname,basename(argv[0])); /* Get `basename $0` */

   while ((param = getopt(argc,argv,"c:efl:psv?")) != EOF)
   {
      switch (param)
      {
         case '?': usage(); /* Bad option */
                   break;
         case 'c': startcol=atoi(optarg);
                   break;
         case 'e': sensitive=1;
                   break;
         case 'f': force=1;
                   break;
         case 'l': startline=atoi(optarg);
                   break;
         case 'p': padit=1;
                   break;
         case 's': word=0;
                   break;
         case 'v': verbose=1;
                   break;
      }
   }

   while (looping)
   {
      if (optind+2>argc) leave("You must supply both the old and new strings");

      oldtmp=strlen(argv[optind]);
      if (!oldtmp) leave("Old string must be non-null");

      oldstrlen[++numstrs]=oldtmp;
      oldstr[numstrs]=(char *)malloc((MALLOCPAR)(oldtmp+1));
      (void)strcpy(oldstr[numstrs],argv[optind++]);

      newtmp=strlen(argv[optind]);
      if (padit && newtmp<oldtmp) alloclen=oldtmp; else alloclen=newtmp;
      newstrlen[numstrs]=alloclen;
      newstr[numstrs]=(char *)malloc((MALLOCPAR)(alloclen+1));
      (void)strcpy(newstr[numstrs],argv[optind++]);
      if (padit)
      {
         while (newtmp<oldtmp) { newstr[numstrs][newtmp++]=' '; }
         newstr[numstrs][newtmp]='\0';
      }

      if ((sensitive && !strcmp(oldstr[numstrs],newstr[numstrs])) ||
           (!sensitive && !strcasecmp(oldstr[numstrs],newstr[numstrs])))
         leave("Old and new strings must be different");

      /* Yes, I could short-cut this, but some lints complain if I do */
      looping=(optind<argc);
      if (looping)
      {
         looping=(!strcmp(argv[optind],"-a") && numstrs<MAXSTRS);
         if (looping) optind++;
      }
   }

   if (verbose)
      (void)fprintf(stderr,"Replacements start from column %d on each line and from line %d onwards\n",startcol,startline);
   startcol--; /* Adjust start column to match char array index */
}

static void read_lines(fd,fdout)
FILE *fd,*fdout;
{
   char thestring[MAXSTRLEN];
   linenum=0; repcount=0;
   while (fgets(thestring,MAXSTRLEN,fd)!=(char *)NULL)
   {
      linenum++;
      thestring[strlen(thestring)-1]='\0';
      if (numstrs==1)
         (void)fprintf(fdout,"%s\n",replace_line(thestring,1));
      else
      {
         int eachstr;
         for (eachstr=1;eachstr<=numstrs;eachstr++)
         {
            /* Not efficient, but hey who's reading this anyway ? */
            char tempstr[MAXSTRLEN];
            (void)strcpy(tempstr,replace_line(thestring,eachstr));
            (void)strcpy(thestring,tempstr);
         }
         (void)fprintf(fdout,"%s\n",thestring);
      }
   }
   if (verbose)
   {
      (void)fprintf(stderr,"%s contained ",intitle);
      plural("line",linenum);
      (void)fprintf(stderr,"and had ");
      plural("string",repcount);
      (void)fprintf(stderr,"replaced\n");
   }
}

static int rename_file(source,dest)
char *source,*dest;
{
    if (rename(source,dest))
    {
         /* Note that the rename function CANNOT rename across filing systems,
            but /bin/mv CAN, so might as well try /bin/mv if rename fails */
         char command[MAXSTRLEN];
         (void)sprintf(command,"/bin/mv -f %s %s 2>/dev/null",source,dest);
         if (system(command)) return(0);
    }
    return(1);
}

static void update_file(oldname,newname,orgname)
char *oldname,*newname,*orgname;
{
   struct stat finfo;
   (void)stat(orgname,&finfo);
   if (rename_file(oldname,newname))
   {
      updated=1;
      if (verbose)
      {
         (void)fprintf(stderr,"%s updated",newname);
         if (force) (void)fprintf(stderr,"\n");
         else (void)fprintf(stderr," (backup with .cln extension)\n");
      }
      (void)chmod(newname,finfo.st_mode);
      (void)chown(newname,finfo.st_uid,finfo.st_gid);
   }
   else (void)fprintf(stderr,"Failed to update %s\n",newname);
}

static void parse_files(argc,argv)
int argc;
char **argv;
{
   FILE *filedes,*fileout;
   for (;optind<argc;optind++)
   {
      (void)strcpy(intitle,argv[optind]);
      if ((filedes=fopen(intitle,"r"))==(FILE *)NULL)
           (void)fprintf(stderr,"Warning: File %s skipped - could not open\n",
                   intitle);
      else
      {
         char tmpname[MAXSTRLEN];
         (void)strcpy(tmpname,mktemp("/tmp/replace_XXXXXX"));
         if ((fileout=fopen(tmpname,"w"))==(FILE *)NULL)
         {
            (void)fprintf(stderr,
                    "Warning: Could not create temp file - skipped %s\n",
                    intitle);
            (void)fclose(filedes);
         }
         else
         {
            read_lines(filedes,fileout); (void)fclose(fileout);
            (void)fclose(filedes);
            updated=0;
            if (repcount)
            {
               if (force) update_file(tmpname,intitle,intitle);
               else
               {
                  char outtitle[MAXSTRLEN];
                  (void)strcpy(outtitle,intitle);
                  (void)strcat(outtitle,".cln");
                  if (rename_file(intitle,outtitle))
                     update_file(tmpname,intitle,outtitle);
               }
            }
            if (!updated) (void)unlink(tmpname);
         }
      }
   }
}

int main(argc,argv)
int argc;
char **argv;
{
   get_options(argc,argv);
   if (optind==argc)
   {
      (void)strcpy(intitle,"<stdin>");
      read_lines(stdin,stdout);
   }
   else parse_files(argc,argv);
   tidy_up();
   return(0);
}
